from rest_framework import serializers

from demo.models import Invoice


class InvoiceSerializer(serializers.ModelSerializer):
    actions = serializers.SerializerMethodField()

    class Meta:
        model = Invoice
        fields = ('id', 'status', 'customer_received', 'is_available', 'actions')
        readonly = ('id', 'status')

    def get_actions(self, instance):
        return sorted(set([transition.action_name for transition in
                           instance.invoice_process.get_available_transitions()]))



from django.test import TestCase
from demo.models import Invoice, Order
from django_logic.transition import Transition


class TransitionTestCase(TestCase):
    def setUp(self) -> None:
        self.invoice = Invoice.objects.create(status='draft')  # uses mixin
        self.order = Order.objects.create(payment_status='paid')  # doesn't use mixin
        self.transition = Transition('test', sources=[], target='cancelled')

    def test_get_hash(self):
        self.assertEqual(self.transition.state.get_hash(self.order, 'payment_status'),
                         'demo-order-payment_status-{}'.format(self.order.pk))
        self.assertEqual(self.transition.state.get_hash(self.invoice, 'status'),
                         'demo-invoice-status-{}'.format(self.invoice.pk))

    def test_get_db_state(self):
        self.assertEqual(self.transition.state.get_db_state(self.invoice, 'status'), 'draft')
        self.assertEqual(self.transition.state.get_db_state(self.order, 'payment_status'), 'paid')
    
    def test_lock(self):
        self.assertFalse(self.transition.state.is_locked(self.invoice, 'status'))
        self.assertFalse(self.transition.state.is_locked(self.order, 'payment_status'))
        self.transition.state.lock(self.invoice, 'status')
        self.transition.state.lock(self.order, 'payment_status')
        self.assertTrue(self.transition.state.is_locked(self.invoice, 'status'))
        self.assertTrue(self.transition.state.is_locked(self.order, 'payment_status'))

        # nothing should happen
        self.transition.state.lock(self.invoice, 'status')
        self.transition.state.lock(self.order, 'payment_status')
        self.assertTrue(self.transition.state.is_locked(self.invoice, 'status'))
        self.assertTrue(self.transition.state.is_locked(self.order, 'payment_status'))

        self.transition.state.unlock(self.invoice, 'status')
        self.transition.state.unlock(self.order, 'payment_status')
        self.assertFalse(self.transition.state.is_locked(self.invoice, 'status'))
        self.assertFalse(self.transition.state.is_locked(self.order, 'payment_status'))

    def test_set_state(self):
        self.transition.state.set_state(self.order, 'payment_status', 'unpaid')
        self.transition.state.set_state(self.invoice, 'status', 'void')
        self.assertEqual(self.order.payment_status, 'unpaid')
        self.assertEqual(self.invoice.status, 'void')
        # make sure it was saved to db
        self.order.refresh_from_db()
        self.invoice.refresh_from_db()
        self.assertEqual(self.order.payment_status, 'unpaid')
        self.assertEqual(self.invoice.status, 'void')

    def test_change_state(self):
        self.transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, self.transition.target)
        self.assertFalse(self.transition.state.is_locked(self.invoice, 'status'))

        self.transition.change_state(self.order, 'payment_status')
        self.assertEqual(self.order.payment_status, self.transition.target)
        self.assertFalse(self.transition.state.is_locked(self.order, 'payment_status'))


def disable_invoice(invoice: Invoice):
    invoice.is_available = False
    invoice.save()


def update_invoice(invoice, is_available, customer_received):
    invoice.is_available = is_available
    invoice.customer_received = customer_received
    invoice.save()


def enable_invoice(invoice: Invoice):
    invoice.is_available = True
    invoice.save()


def fail_invoice(invoice: Invoice):
    raise Exception


def receive_invoice(invoice: Invoice):
    invoice.customer_received = True
    invoice.save()


class TransitionSideEffectsTestCase(TestCase):
    def setUp(self) -> None:
        self.invoice = Invoice.objects.create(status='draft')

    def test_one_side_effect(self):
        transition = Transition('test', sources=[], target='cancelled', side_effects=[disable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, transition.target)
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_many_side_effects(self):
        transition = Transition('test', sources=[], target='cancelled',
                                side_effects=[disable_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, transition.target)
        self.assertTrue(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_failure_during_side_effect(self):
        transition = Transition('test', sources=[], target='cancelled',
                                side_effects=[disable_invoice, fail_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'draft')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_failure_during_side_effect_with_failed_state(self):
        transition = Transition('test', sources=[], target='cancelled', failed_state='failed',
                                side_effects=[disable_invoice, fail_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'failed')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_side_effect_with_parameters(self):
        update_invoice(self.invoice, is_available=True, customer_received=True)
        transition = Transition('test', sources=[], target='cancelled', failed_state='failed',
                                side_effects=[update_invoice])
        self.invoice.refresh_from_db()
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)
        transition.change_state(self.invoice, 'status', is_available=False, customer_received=False)
        self.invoice.refresh_from_db()
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)


class TransitionCallbacksTestCase(TestCase):
    def setUp(self) -> None:
        self.invoice = Invoice.objects.create(status='draft')

    def test_one_callback(self):
        transition = Transition('test', sources=[], target='cancelled', callbacks=[disable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, transition.target)
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_many_callbacks(self):
        transition = Transition('test', sources=[], target='cancelled',
                                callbacks=[disable_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, transition.target)
        self.assertTrue(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_failure_during_callbacks(self):
        transition = Transition('test', sources=[], target='cancelled',
                                callbacks=[disable_invoice, fail_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'cancelled')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_failure_during_callbacks_with_failed_state(self):
        transition = Transition('test', sources=[], target='cancelled', failed_state='failed',
                                side_effects=[disable_invoice, fail_invoice, enable_invoice])
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'failed')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_callbacks_with_parameters(self):
        update_invoice(self.invoice, is_available=True, customer_received=True)
        transition = Transition('test', sources=[], target='cancelled', failed_state='failed',
                                callbacks=[update_invoice])
        self.invoice.refresh_from_db()
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)
        transition.change_state(self.invoice, 'status', is_available=False, customer_received=False)
        self.invoice.refresh_from_db()
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)


class TransitionFailureCallbacksTestCase(TestCase):
    def setUp(self) -> None:
        self.invoice = Invoice.objects.create(status='draft')

    def test_one_callback(self):
        transition = Transition('test', sources=[], target='success', side_effects=[fail_invoice],
                                failure_callbacks=[disable_invoice], failed_state='failed')
        self.assertTrue(self.invoice.is_available)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'failed')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_many_callback(self):
        transition = Transition('test', sources=[], target='success', side_effects=[fail_invoice],
                                failure_callbacks=[disable_invoice, receive_invoice], failed_state='failed')
        self.assertTrue(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)
        transition.change_state(self.invoice, 'status')
        self.assertEqual(self.invoice.status, 'failed')
        self.assertFalse(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)
        self.assertFalse(transition.state.is_locked(self.invoice, 'status'))

    def test_callbacks_with_parameters(self):
        update_invoice(self.invoice, is_available=True, customer_received=True)
        transition = Transition('test', sources=[], target='success', failed_state='failed',
                                side_effects=[fail_invoice], failure_callbacks=[update_invoice])
        self.invoice.refresh_from_db()
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)
        transition.change_state(self.invoice, 'status', is_available=False, customer_received=False)
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'failed')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)


from django.test import TestCase

from demo.models import Invoice

from django_logic import Process, Transition


class User:
    is_allowed = True


def allow(instance, user):
    return user is not None and user.is_allowed


def disallow(instance, user):
    return False


def is_editable(instance):
    return not instance.customer_received


def not_available(instance):
    return False


class ValidateProcessTestCase(TestCase):
    def setUp(self) -> None:
        self.user = User()

    def test_pure_process(self):
        class MyProcess(Process):
            pass

        process = MyProcess(field_name='status')
        self.assertTrue(process.validate())
        self.assertTrue(process.validate(self.user))

    def test_empty_permissions(self):
        class MyProcess(Process):
            permissions = []

        self.assertTrue(MyProcess('state').validate())
        self.assertTrue(MyProcess('state').validate(self.user))

    def test_permissions_successfully(self):
        class MyProcess(Process):
            permissions = [allow]

        self.assertFalse(MyProcess('state').validate())
        self.assertTrue(MyProcess('state').validate(self.user))

    def test_permission_fail(self):
        self.user.is_allowed = False

        class MyProcess(Process):
            permissions = [allow]

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

        class AnotherProcess(Process):
            permissions = [allow, disallow]

        process = AnotherProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

    def test_empty_conditions(self):
        class MyProcess(Process):
            conditions = []

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertTrue(process.validate(self.user))

    def test_conditions_successfully(self):
        class MyProcess(Process):
            conditions = [is_editable]

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertTrue(process.validate())
        self.assertTrue(process.validate(self.user))

    def test_conditions_fail(self):
        class MyProcess(Process):
            conditions = [not_available]

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))

        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

        class AnotherProcess(Process):
            conditions = [is_editable]

        instance = Invoice(status='draft')
        instance.customer_received = True
        process = AnotherProcess(field_name='status', instance=instance)
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

    def test_permissions_and_conditions_successfully(self):
        class MyProcess(Process):
            permissions = [allow]
            conditions = [is_editable]

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertTrue(process.validate(self.user))

    def test_permissions_and_conditions_fail(self):
        class MyProcess(Process):
            permissions = [allow, disallow]
            conditions = [is_editable]

        process = MyProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

        class AnotherProcess(Process):
            permissions = [allow]
            conditions = [is_editable, not_available]

        process = AnotherProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))

        class FinalProcess(Process):
            permissions = [allow, disallow]
            conditions = [is_editable, not_available]

        process = FinalProcess(field_name='status', instance=Invoice(status='draft'))
        self.assertFalse(process.validate())
        self.assertFalse(process.validate(self.user))


class GetAvailableTransitionsTestCase(TestCase):
    def setUp(self) -> None:
        self.user = User()

    def test_pure_process(self):
        class ChildProcess(Process):
            pass

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [])

    def test_process(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [transition1])

        process = ChildProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [transition2])

        process = ChildProcess(instance=Invoice.objects.create(status='closed'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [])

    def test_process_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            conditions = [not_available]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [])

        process = ChildProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [])

        process = ChildProcess(instance=Invoice.objects.create(status='closed'), field_name='status')
        self.assertEqual(list(process.get_available_transitions()), [])

    def test_conditions_and_permissions_successfully(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            conditions = [is_editable]
            permissions = [allow]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        process = ChildProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition2])

        process = ChildProcess(instance=Invoice.objects.create(status='closed'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

    def test_conditions_and_permissions_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            conditions = [is_editable]
            permissions = [disallow]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

        process = ChildProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

        process = ChildProcess(instance=Invoice.objects.create(status='closed'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

    def test_nested_process_permissions_successfully(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            permissions = [allow]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class ParentProcess(Process):
            permissions = [allow]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class GrandParentProcess(Process):
            permissions = [allow]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

    def test_nested_process_permissions_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            permissions = [disallow]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

        class ParentProcess(Process):
            permissions = [allow]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

        class GrandParentProcess(Process):
            permissions = [allow]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

    def test_nested_process_conditions_successfully(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            conditions = [is_editable]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class ParentProcess(Process):
            conditions = [is_editable]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class GrandParentProcess(Process):
            conditions = [is_editable]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

    def test_nested_process_conditions_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            conditions = [is_editable]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class ParentProcess(Process):
            conditions = [not_available]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

        class GrandParentProcess(Process):
            conditions = [is_editable]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

    def test_nested_process_successfully(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class ParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class GrandParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

    def test_nested_process_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')

        class ChildProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            transitions = [transition1, transition2]

        process = ChildProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class ParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ChildProcess,)

        process = ParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [transition1])

        class GrandParentProcess(Process):
            permissions = [allow]
            conditions = [not_available]
            nested_processes = (ParentProcess,)

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        self.assertEqual(list(process.get_available_transitions(self.user)), [])

    def test_nested_process_with_nested_transitions_successfully(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')
        transition3 = Transition('action', sources=['draft'], target='approved')
        transition4 = Transition('action', sources=['done'], target='closed')
        transition5 = Transition('action', sources=['draft'], target='declined')

        class ChildProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            transitions = [transition1, transition2]

        class ParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ChildProcess,)
            transitions = [transition3, transition4]

        class GrandParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ParentProcess,)

            transitions = [transition5]

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')

        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition1, transition3, transition5])

        process = GrandParentProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition2, transition4])

    def test_nested_process_with_nested_transitions_fail(self):
        transition1 = Transition('action', sources=['draft'], target='done')
        transition2 = Transition('action', sources=['done'], target='closed')
        transition3 = Transition('action', sources=['draft'], target='approved')
        transition4 = Transition('action', sources=['done'], target='closed')
        transition5 = Transition('action', sources=['draft'], target='declined')

        class ChildProcess(Process):
            permissions = [disallow]
            conditions = [is_editable]
            transitions = [transition1, transition2]

        class ParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ChildProcess,)
            transitions = [transition3, transition4]

        class GrandParentProcess(Process):
            permissions = [allow]
            conditions = [is_editable]
            nested_processes = (ParentProcess,)

            transitions = [transition5]

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition3, transition5])

        process = GrandParentProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition4])

    def test_nested_process_with_nested_transitions_conditions_and_permissions_successfully(self):
        transition1 = Transition('action', permissions=[allow], conditions=[is_editable],
                                 sources=['draft'],
                                 target='done')
        transition2 = Transition('action', permissions=[allow], conditions=[is_editable],
                                 sources=['done'],
                                 target='closed')
        transition3 = Transition('action',
                                 permissions=[allow],
                                 conditions=[is_editable],
                                 sources=['draft'],
                                 target='approved')
        transition4 = Transition('action',
                                 permissions=[allow],
                                 conditions=[is_editable],
                                 sources=['done'],
                                 target='closed')
        transition5 = Transition('action',
                                 permissions=[allow],
                                 conditions=[is_editable],
                                 sources=['draft'],
                                 target='declined')

        class ChildProcess(Process):
            transitions = [transition1, transition2]

        class ParentProcess(Process):
            nested_processes = (ChildProcess,)
            transitions = [transition3, transition4]

        class GrandParentProcess(Process):
            nested_processes = (ParentProcess,)

            transitions = [transition5]

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition1, transition3, transition5])

        process = GrandParentProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition2, transition4])

    def test_nested_process_with_nested_transitions_conditions_and_permissions_fail(self):
        transition1 = Transition('action',
                                 permissions=[allow],
                                 conditions=[is_editable],
                                 sources=['draft'],
                                 target='done')
        transition2 = Transition('action',
                                 permissions=[disallow],
                                 conditions=[is_editable],
                                 sources=['done'],
                                 target='closed')
        transition3 = Transition('action',
                                 permissions=[allow],
                                 conditions=[not_available],
                                 sources=['draft'],
                                 target='approved')
        transition4 = Transition('action',
                                 permissions=[allow],
                                 conditions=[is_editable],
                                 sources=['done'],
                                 target='closed')
        transition5 = Transition('action',
                                 permissions=[disallow],
                                 conditions=[not_available],
                                 sources=['draft'],
                                 target='declined')

        class ChildProcess(Process):
            transitions = [transition1, transition2]

        class ParentProcess(Process):
            nested_processes = (ChildProcess,)
            transitions = [transition3, transition4]

        class GrandParentProcess(Process):
            nested_processes = (ParentProcess,)

            transitions = [transition5]

        process = GrandParentProcess(instance=Invoice.objects.create(status='draft'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition1])

        process = GrandParentProcess(instance=Invoice.objects.create(status='done'), field_name='status')
        for transition in process.get_available_transitions(self.user):
            self.assertIn(transition, [transition4])


def disable_invoice(invoice: Invoice):
    invoice.is_available = False
    invoice.customer_received = False
    invoice.save()


def update_invoice(invoice, is_available, customer_received):
    invoice.is_available = is_available
    invoice.customer_received = customer_received
    invoice.save()


def enable_invoice(invoice: Invoice):
    invoice.is_available = True
    invoice.save()


def fail_invoice(invoice: Invoice):
    raise Exception


class ApplyTransitionTestCase(TestCase):
    def setUp(self) -> None:
        self.user = User()
        self.invoice = Invoice.objects.create(status='draft')

    def test_simple_transition(self):
        class TestProcess(Process):
            transitions = [
                Transition('cancel', sources=['draft', ], target='cancelled')
            ]

        process = TestProcess(instance=self.invoice, field_name='status')
        process.cancel()
        self.assertEqual(self.invoice.status, 'cancelled')

    def test_several_transitions(self):
        class TestProcess(Process):
            transitions = [
                Transition('cancel', sources=['draft', ], target='cancelled'),
                Transition('undo', sources=['cancelled'], target='draft')
            ]

        process = TestProcess(instance=self.invoice, field_name='status')
        process.cancel()
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'cancelled')
        process.undo()
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'draft')

    def test_transition_with_side_effect(self):
        class TestProcess(Process):
            transitions = [
                Transition('cancel', sources=['draft', ], target='cancelled', side_effects=[disable_invoice]),
                Transition('undo', sources=['cancelled'], target='draft', side_effects=[update_invoice])
            ]
        self.assertTrue(self.invoice.is_available)
        process = TestProcess(instance=self.invoice, field_name='status')
        process.cancel()
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'cancelled')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)
        process.undo(is_available=True, customer_received=True)
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'draft')
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)

    def test_transition_with_callbacks(self):
        class TestProcess(Process):
            transitions = [
                Transition('cancel', sources=['draft', ], target='cancelled', callbacks=[disable_invoice]),
                Transition('undo', sources=['cancelled'], target='draft', callbacks=[update_invoice])
            ]
        self.assertTrue(self.invoice.is_available)
        process = TestProcess(instance=self.invoice, field_name='status')
        process.cancel()
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'cancelled')
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)
        process.undo(is_available=True, customer_received=True)
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'draft')
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)

    def test_transition_with_failure_callbacks(self):
        class TestProcess(Process):
            transitions = [
                Transition('cancel', sources=['draft', ], target='cancelled', callbacks=[disable_invoice]),
                Transition('undo', sources=['draft'], target='draft', side_effects=[fail_invoice], failed_state='failed', failure_callbacks=[update_invoice])
            ]
        update_invoice(self.invoice, is_available=False, customer_received=False)
        self.assertFalse(self.invoice.is_available)
        self.assertFalse(self.invoice.customer_received)
        self.assertEqual(self.invoice.status, 'draft')
        process = TestProcess(instance=self.invoice, field_name='status')
        process.undo(is_available=True, customer_received=True)
        self.invoice.refresh_from_db()
        self.assertEqual(self.invoice.status, 'failed')
        self.assertTrue(self.invoice.is_available)
        self.assertTrue(self.invoice.customer_received)



from django_logic import Process, Transition


class InvoiceProcess(Process):
    process_name = 'invoice_process'
    states = (
        ('draft', 'Draft'),
        ('paid', 'Paid'),
        ('void', 'Void'),
        ('sent', 'Sent'),
        ('failed', 'Failed'),
    )

    transitions = [
        Transition(
            action_name='approve',
            sources=['draft'],
            target='approved'
        ),
        Transition(
            action_name='void',
            sources=['draft', 'paid'],
            target='voided'
        ),
    ]


from rest_framework import routers

from demo.viewsets import InvoiceViewSet

app_name = 'api'

router = routers.DefaultRouter()
router.register(r'invoices', InvoiceViewSet, basename='invoice')
urlpatterns = router.urls


import re

from graphviz import Digraph


def get_object_id(obj):
    return str(id(obj))


def get_conditions_id(obj):
    return '{}|conditions'.format(id(obj))


def get_readable_process_name(process) -> str:
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1 \2', str(process.__name__))
    return re.sub('([a-z0-9])([A-Z])', r'\1 \2', s1)


def get_target_states(process) -> set:
    states = set()
    for transition in process.transitions:
        states.add(transition.target)
        if transition.in_progress_state:
            states.add(transition.in_progress_state)
    return states


def get_all_target_states(process) -> set:
    states = get_target_states(process)
    for sub_process in process.nested_processes:
        states |= get_all_target_states(sub_process)
    return states


def get_all_states(process) -> set:
    states = set()
    for transition in process.transitions:
        states.add(transition.target)
        if transition.in_progress_state:
            states.add(transition.in_progress_state)
        states |= set(transition.sources)
        for sub_process in process.nested_processes:
            states |= get_all_states(sub_process)
    return states


def annotate_nodes(process):
    """
      This function annotate node names and nodes into two dicts:
      - node_names contains either process or transition unique node name.
      - nodes contains the information of the given process
      :param process: Process class
      It should return a directed graph where every node has unique name,
      nodes could be connected by arrows.

      It assigns a state to a node only and only if the state hasn't been used before in the graph.
      In other words, a state is always on an upper level before it's used. It helps to display sub graphs
      around that state, which provides better understanding how the process structured.

      Supported the following types of nodes:
      - process
      - process_conditions
      - transition
      - transition_conditions
      - state

      and list of paths - from and to

      parameters:
      - node_name - unique node name. It needs to know exact node path,
          as there is no unique identifier between transitions, processes and other nodes.
          So, the node name is combined through the graph path.

      - name - displayed name
      - type - could be: process, process_conditions, transition, transition_conditions

      # Example:
      nodes = {
          'name': 'Main process',
          'type': 'process',
          'nodes': [
              {
                  'name': 'my unique name of the node',
                  'nodes': [
                      {
                          'type': 'transition',
                          'name': 'my transition'
                      },
                      {
                          'type': 'condition',
                          'name': 'my condition'
                      }
                  ],
                  'type': 'process'
              },
              {
                  'name': 'my state',
                  'type': 'state',
              }
          ]
      }
      """
    used_states = set()

    def annotate_process_nodes(process, node_name):
        node_name += get_readable_process_name(process) + '|'
        # process
        node = {
            'id': get_object_id(process),
            'name': get_readable_process_name(process),
            'type': 'process',
            'nodes': []
        }

        # process conditions
        if process.conditions:
            node['nodes'].append({
                'id': get_conditions_id(process),
                'name': '\n'.join([condition.__name__ for condition in process.conditions]),
                'type': 'process_conditions',
            })

        # transitions
        for transition in process.transitions:
            node['nodes'].append({
                'id': get_object_id(transition),
                'name': transition.action_name,
                'type': 'transition',
            })
            # transition conditions
            if transition.conditions.commands:
                node['nodes'].append({
                    'id': get_conditions_id(transition),
                    'name': '\n'.join([condition.__name__ for condition in transition.conditions.commands]),
                    'type': 'transition_conditions',
                })

        # it finds all intersections between the current process' states and its sub process' states
        states = get_target_states(process)
        for sub_process1 in process.nested_processes:
            for sub_process2 in process.nested_processes:
                if sub_process1 != sub_process2:
                    states |= (get_all_target_states(sub_process1) &
                               get_all_target_states(sub_process2))

        # it should assign all intersect states excluding used states before
        for state in states - used_states:
            node['nodes'].append({
                'id': state,
                'name': state,
                'type': 'state',
            })
            used_states.add(state)

        for sub_process in process.nested_processes:
            node['nodes'].append(annotate_process_nodes(sub_process, node_name))
        return node

    node = annotate_process_nodes(process, node_name='')
    for state in get_all_states(process) - used_states:
        node['nodes'].append({
            'id': state,
            'name': state,
            'type': 'state',
        })

    return node


def fsm_paths(process, state):
    paths = set()
    visited_state = []

    def add_path(target, source):
        """
        the path given from the target to the source, but it should swap when added
        """
        paths.add((source, target))

    def get_available_transitions(process_class, state):
        for transition in process_class.transitions:
            if state in transition.sources:
                yield transition
        for sub_process in process_class.nested_processes:
            for transition in get_available_transitions(sub_process, state):
                yield transition

    def dfs(current_state):
        """
        This function goes from "the bottom to the top" through the available transitions and
         consolidate the path from transition to the state,
         including transition.

        target state <- transition <- transition conditions <- current state

        Where the path is a tuple of unique id.
        For example:
        [
            ('target state', '12312312'),  # target state <- transition
            ('12312312', '12312312|conditions'),  # transition <- transition conditions
            ('12312312|conditions', 'current state'),  #  transition conditions <- current state
        ]
        :param current_state:
        :return:
        """
        visited_state.append(current_state)

        for transition in get_available_transitions(process, current_state):
            current = get_object_id(transition)
            add_path(transition.target, current)

            if transition.conditions.commands:
                target = get_conditions_id(transition)
                add_path(current, target)
                current = target

            add_path(current, current_state)

            if transition.target not in visited_state:
                dfs(transition.target)

    dfs(state)
    return paths


def get_graph_from_node(main_node, paths, skip_main_process=False):
    def draw_node(node, graph):
        if node['type'] == 'process':
            graph.attr(label=node['name'])
            return

        if node['type'] == 'process_conditions':
            graph.attr('node', style='filled', fillcolor='white', shape='diamond')

        if node['type'] == 'transition':
            graph.attr('node', style='filled', fillcolor='lightgrey', shape='record')

        if node['type'] == 'transition_conditions':
            graph.attr('node', style='filled', fillcolor='lightgrey', shape='diamond')

        if node['type'] == 'state':
            graph.attr('node', style='filled', fillcolor='white', shape='oval')

        graph.node(name=node['id'], label=node['name'])

    def draw_process(node, graph):
        """
        :param node: node is process
        :param graph:
        :return:
        """
        for node in node['nodes']:
            if node['type'] == 'process' and node['nodes']:
                # for every process we should draw a new cluster
                with graph.subgraph(name="cluster_{}".format(node['id'])) as subgraph:
                    draw_process(node, subgraph)
                    draw_node(node, subgraph)
            else:
                draw_node(node, graph)

    def draw_edges(graph):
        for from_node, to_node in paths:
            graph.edge(from_node, to_node)

    engine = 'fdp'
    digraph = Digraph(main_node['name'], filename=main_node['name'],
                      engine=engine, node_attr={'shape': 'record'})

    if not skip_main_process:
        main_node = dict(nodes=[main_node], type='process')
    draw_process(main_node, digraph)
    draw_edges(digraph)

    digraph.attr(overlap='false')
    return digraph


def get_graph_from_process(process_class, state, skip_main_process=False):
    node = annotate_nodes(process_class)
    paths = fsm_paths(process_class, state)
    graph = get_graph_from_node(node, paths, skip_main_process)
    return graph


def display_process(process_class, state, skip_main_process=False):
    graph = get_graph_from_process(process_class, state, skip_main_process)
    try:
        graph.view()
    except Exception as ex:
        print(ex.stderr)



import logging
from functools import partial

from django_logic.commands import Conditions, Permissions
from django_logic.exceptions import ManyTransitions, TransitionNotAllowed
from django_logic.state import State

logger = logging.getLogger(__name__)


class Process(object):
    """
    Process should be explicitly defined as a class and used as an object.
    - process name
    - nested states
    - contains either transitions and processes
    - transitions defined as parameters of the class
    - processes should be defined in the list
    - validate - conditions and permissions of the process affects all transitions/processes inside
    - has methods like get_all_available_transitions, etc
    """
    nested_processes = []
    transitions = []
    conditions = []
    permissions = []
    conditions_class = Conditions
    permissions_class = Permissions
    process_name = 'process'

    def __init__(self, field_name: str, instance=None):
        """
        :param field_name:
        """
        self.field_name = field_name
        self.instance = instance

    def __getattr__(self, item):
        transitions = list(self.get_available_transitions(action_name=item))

        if len(transitions) == 1:
            return partial(transitions[0].change_state,
                           instance=self.instance,
                           field_name=self.field_name)

        # This exceptions should be handled otherwise it will be very annoying
        elif transitions:
            raise ManyTransitions("There are several transitions available")
        else:
            # TODO: transition not available
            raise TransitionNotAllowed('Transition not allowed')

    def validate(self, user=None) -> bool:
        """
        It validates this process to meet conditions and pass permissions
        :param user: any object used to pass permissions
        :return: True or False
        """
        permissions = self.permissions_class(commands=self.permissions)
        conditions = self.conditions_class(commands=self.conditions)
        return (permissions.execute(self.instance, user) and
                conditions.execute(self.instance))

    def get_available_transitions(self, user=None, action_name=None):
        """
        It returns all available transition which meet conditions and pass permissions.
        Including nested processes.
        :param action_name:
        :param user: any object which used to validate permissions
        :return: yield `django_logic.Transition`
        """
        if not self.validate(user):
            return

        state = State().get_db_state(self.instance, self.field_name)
        for transition in self.transitions:
            if action_name is not None and transition.action_name != action_name:
                continue

            if state in transition.sources and transition.validate(self.instance,
                                                                   self.field_name,
                                                                   user):
                yield transition

        for sub_process_class in self.nested_processes:
            sub_process = sub_process_class(instance=self.instance, field_name=self.field_name)
            for transition in sub_process.get_available_transitions(user=user,
                                                                    action_name=action_name):
                yield transition


class ProcessManager:
    @classmethod
    def bind_state_fields(cls, **kwargs):
        def make_process_getter(field_name, field_class):
            return lambda self: field_class(field_name=field_name, instance=self)

        parameters = {'state_fields': []}
        for state_field, process_class in kwargs.items():
            if not issubclass(process_class, Process):
                raise TypeError('Must be a sub class of Process')
            parameters[process_class.process_name] = property(make_process_getter(state_field, process_class))
            parameters['state_fields'].append(state_field)
        return type('Process', (cls, ), parameters)

    @property
    def non_state_fields(self):
        """
        Returns list of object's non-state fields.
        """
        field_names = set()
        for field in self._meta.fields:
            if not field.primary_key and not field.name in self.state_fields:
                field_names.add(field.name)

                if field.name != field.attname:
                    field_names.add(field.attname)
        return field_names

    def save(self, *args, **kwargs):
        """
        It saves all non-state fields by default.
        State fields can be saved if explicitly passed in 'update_fields' kwarg.
        """
        if self.id is not None and 'update_fields' not in kwargs:
            kwargs['update_fields'] = self.non_state_fields
        super().save(*args, **kwargs)



from django.test import TestCase

from demo.models import Invoice
from demo.process import InvoiceProcess


class InvoiceProcessTestCase(TestCase):
    def setUp(self):
        self.process_class = InvoiceProcess

    def test_process_class_method(self):
        self.assertEqual(self.process_class.process_name, 'invoice_process')

    def test_invoice_process(self):
        invoice = Invoice.objects.create(status='draft')
        invoice.invoice_process.approve()
        invoice.refresh_from_db()
        self.assertEqual(invoice.status, 'approved')

    def test_invoice_save(self):
        invoice = Invoice.objects.create(status='draft')
        invoice.status = 'paid'
        invoice.save()
        invoice.refresh_from_db()
        self.assertEqual(invoice.status, 'draft')
        invoice.invoice_process.approve()
        invoice.refresh_from_db()
        self.assertEqual(invoice.status, 'approved')



from rest_framework import viewsets

from demo.models import Invoice
from demo.serializers import InvoiceSerializer


class InvoiceViewSet(viewsets.ModelViewSet):
    queryset = Invoice.objects.all()
    serializer_class = InvoiceSerializer



from django.core.cache import cache


class State(object):
    @staticmethod
    def get_db_state(instance, field_name):
        """
        Fetches state directly from db instead of model instance.
        """
        return instance._meta.model.objects.values_list(field_name, flat=True).get(pk=instance.id)

    @staticmethod
    def set_state(instance, field_name, state):
        """
        Sets intermediate state to instance's field until transition is over.
        """
        # TODO: how would it work if it's used within another transaction?
        instance._meta.model.objects.filter(pk=instance.id).update(**{field_name: state})
        instance.refresh_from_db()

    def get_hash(self, instance, field_name):
        # TODO: https://github.com/Borderless360/django-logic/issues/3
        return "{}-{}-{}-{}".format(instance._meta.app_label,
                                    instance._meta.model_name,
                                    field_name,
                                    instance.pk)

    def lock(self, instance, field_name: str):
        cache.set(self.get_hash(instance, field_name), True)

    def unlock(self, instance, field_name: str):
        cache.delete(self.get_hash(instance, field_name))

    def is_locked(self, instance, field_name: str):
        return cache.get(self.get_hash(instance, field_name)) or False


from django.test import TestCase

from django_logic.process import ProcessManager, Process
from django_logic.transition import Transition


class FirstProcess(Process):
    process_name = 'first_process'
    transitions = [
        Transition(action_name='transition1', sources=['state1'], target='state3')
    ]


class SecondProcess(Process):
    process_name = 'second_process'
    transitions = [
        Transition(action_name='transition2', sources=['state2'], target='state3')
    ]


class ProcessManagerTestCase(TestCase):
    def test_processes_bound_correctly(self):
        bind_class = ProcessManager.bind_state_fields(first_state=FirstProcess, second_state=SecondProcess)
        bind_class_obj = bind_class()
        self.assertTrue(hasattr(bind_class_obj, 'first_process'))
        self.assertTrue(isinstance(bind_class_obj.first_process, FirstProcess))
        self.assertTrue(hasattr(bind_class_obj, 'second_process'))
        self.assertTrue(isinstance(bind_class_obj.second_process, SecondProcess))



# Generated by Django 2.2.7 on 2019-12-06 00:17

from django.db import migrations, models
import django_logic.process


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Invoice',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.CharField(blank=True, choices=[('draft', 'Draft'), ('paid', 'Paid'), ('void', 'Void')], default='draft', max_length=16, null=True)),
                ('customer_received', models.BooleanField(default=False)),
                ('is_available', models.BooleanField(default=True)),
            ],
            bases=(django_logic.process.Process, models.Model),
        ),
        migrations.CreateModel(
            name='Order',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('payment_status', models.CharField(blank=True, max_length=16)),
            ],
        ),
    ]



import os
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'django_logic'

PROJECT_APPS = [
    'django_logic',
    'demo',
]

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
] + PROJECT_APPS

try:
    import rest_framework
except ImportError:
    pass
else:
    INSTALLED_APPS += ['rest_framework']
    REST_FRAMEWORK = {
        'DEFAULT_PERMISSION_CLASSES': [
            'rest_framework.permissions.AllowAny'
        ]
    }
ROOT_URLCONF = 'tests.urls'

MIDDLEWARE = []

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'django_logic',
    }
}
#
# MIGRATION_MODULES = {
#     'auth': None,
#     'contenttypes': None,
# }


from django.urls import include, path, reverse
from rest_framework.test import APITestCase
from rest_framework import status

from demo.models import Invoice


class InvoiceAPITestCase(APITestCase):
    def test_create_invoice(self):
        url = reverse('demo:invoice-list')
        response = self.client.post(url, data={}, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        invoice = Invoice.objects.last()
        self.assertEqual(response.json(), {'id': invoice.pk,
                                           'status': 'draft',
                                           'actions': ['approve', 'void'],
                                           'customer_received': False,
                                           'is_available': True})

    def test_get_invoices(self):
        url = reverse('demo:invoice-list')
        invoice = Invoice.objects.create(status='draft')
        response = self.client.get(url)
        self.assertEqual(response.json(),
                         [{'id': invoice.pk,
                           'status': 'draft',
                           'actions': ['approve', 'void'],
                           'customer_received': False,
                           'is_available': True}])



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys

PROJECT_ROOT = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))
sys.path.insert(0, PROJECT_ROOT)


def main():
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tests.settings')
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="django-logic",
    version="0.0.8",
    author="Emil Balashov",
    author_email="emil@borderless360.com",
    description="Django Logic - easy way to implement state-based business logic",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Borderless360/django-logic",
    keywords="django",
    packages=setuptools.find_packages(),
    include_package_data=True,
    zip_safe=False,
    license='MIT License',
    platforms=['any'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.6',
)


from django_logic.commands import SideEffects, Callbacks, Permissions, Conditions
from django_logic.exceptions import TransitionNotAllowed
from django_logic.state import State


class Transition(object):
    """
    Transition could be defined as a class and used as an object
    - action name
    - transitions name
    - target
    - it changes the the state of the object from source to target by triggering available action via transition name
    - validation if the action is available throughout permissions and conditions
    - run side effects and call backs
    """
    side_effects_class = SideEffects
    callbacks_class = Callbacks
    failure_callbacks_class = Callbacks
    permissions_class = Permissions
    conditions_class = Conditions
    state = State()

    def __init__(self, action_name: str, sources: list, target: str, **kwargs):
        self.action_name = action_name
        self.target = target
        self.sources = sources
        self.in_progress_state = kwargs.get('in_progress_state')
        self.failed_state = kwargs.get('failed_state')
        self.failure_callbacks = self.failure_callbacks_class(kwargs.get('failure_callbacks', []), transition=self)
        self.side_effects = self.side_effects_class(kwargs.get('side_effects', []), transition=self)
        self.callbacks = self.callbacks_class(kwargs.get('callbacks', []), transition=self)
        self.permissions = self.permissions_class(kwargs.get('permissions', []), transition=self)
        self.conditions = self.conditions_class(kwargs.get('conditions', []), transition=self)

    def __str__(self):
        return "Transition: {} to {}".format(self.action_name, self.target)

    def validate(self, instance: any, field_name: str, user=None) -> bool:
        """
        It validates this process to meet conditions and pass permissions
        :param field_name:
        :param instance: any instance used to meet conditions
        :param user: any object used to pass permissions
        :return: True or False
        """
        return (not self.state.is_locked(instance, field_name) and
                self.permissions.execute(instance, user) and
                self.conditions.execute(instance))

    def change_state(self, instance, field_name, **kwargs):
        """
        This method changes a state of the provided instance and file name by the following algorithm:
        - Lock state
        - Change state to `in progress` if such exists
        - Run side effects which should run `complete_transition` in case of success
        or `fail_transition` in case of failure.
        :param instance: any
        :param field_name: str
        """
        if self.state.is_locked(instance, field_name):
            raise TransitionNotAllowed("State is locked")

        self.state.lock(instance, field_name)
        if self.in_progress_state:
            self.state.set_state(instance, field_name, self.in_progress_state)
        self.side_effects.execute(instance, field_name, **kwargs)

    def complete_transition(self, instance, field_name, **kwargs):
        """
        It completes the transition process for provided instance and filed name.
        The instance will be unlocked and callbacks exc
        :param instance:
        :param field_name:
        :return:
        """
        self.state.set_state(instance, field_name, self.target)
        self.state.unlock(instance, field_name)
        self.callbacks.execute(instance, field_name, **kwargs)
    
    def fail_transition(self, instance, field_name, **kwargs):
        """
        It triggers fail transition in case of any failure during the side effects execution.
        :param instance: any
        :param field_name: str
        """
        if self.failed_state:
            self.state.set_state(instance, field_name, self.failed_state)
        self.state.unlock(instance, field_name)
        self.failure_callbacks.execute(instance, field_name, **kwargs)


class BaseCommand(object):
    """
    Command
    """
    def __init__(self, commands=None, transition=None):
        self._commands = commands or []
        self._transition = transition

    @property
    def commands(self):
        return self._commands

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class Conditions(BaseCommand):
    def execute(self, instance: any, **kwargs):
        return all(command(instance, **kwargs) for command in self._commands)


class Permissions(BaseCommand):
    def execute(self, instance: any, user: any, **kwargs):
        return all(command(instance,  user, **kwargs) for command in self._commands)


class SideEffects(BaseCommand):
    def execute(self, instance: any, field_name: str, **kwargs):
        try:
            for command in self._commands:
                command(instance, **kwargs)
        except Exception:
            self._transition.fail_transition(instance, field_name, **kwargs)
        else:
            self._transition.complete_transition(instance, field_name, **kwargs)


class Callbacks(BaseCommand):
    def execute(self, instance: any, field_name: str, **kwargs):
        try:
            for command in self.commands:
                command(instance, **kwargs)
        except Exception:
            # TODO: handle exception
            pass


from django.db import models

from demo.process import InvoiceProcess

from django_logic.process import ProcessManager


class Invoice(ProcessManager.bind_state_fields(status=InvoiceProcess), models.Model):
    status = models.CharField(choices=InvoiceProcess.states, default='draft', max_length=16, blank=True)
    customer_received = models.BooleanField(default=False)
    is_available = models.BooleanField(default=True)

    def __str__(self):
        return self.status


class Order(models.Model):
    payment_status = models.CharField(max_length=16, blank=True)
